"""
Set of tests for adaptive exponential integrate-and-fire mean-field model.
"""

import unittest

import numba
import numpy as np
import pytest
import xarray as xr
from jitcdde import jitcdde_input
from neurolib.models.aln import ALNModel
from neurolib.models.multimodel.builder.aln import (
    ALN_EXC_DEFAULT_PARAMS,
    ALN_INH_DEFAULT_PARAMS,
    ALN_NODE_DEFAULT_CONNECTIVITY,
    ALNNetwork,
    ALNNode,
    ExcitatoryALNMass,
    InhibitoryALNMass,
    _get_interpolation_values,
    _table_lookup,
)
from neurolib.models.multimodel.builder.base.constants import EXC
from neurolib.utils.stimulus import ZeroInput

# these keys do not test since they are rescaled on the go
PARAMS_NOT_TEST_KEYS = ["c_gl", "taum", "input_0"]


def _strip_keys(dict_test, strip_keys=PARAMS_NOT_TEST_KEYS):
    return {k: v for k, v in dict_test.items() if k not in strip_keys}


SEED = 42
DURATION = 100.0
DT = 0.01
CORR_THRESHOLD = 0.9
NEUROLIB_VARIABLES_TO_TEST = [("r_mean_EXC", "rates_exc"), ("r_mean_INH", "rates_inh")]

# dictionary as backend name: format in which the noise is passed
BACKENDS_TO_TEST = {
    "jitcdde": lambda x, d, dt: x.as_cubic_splines(d, dt),
    "numba": lambda x, d, dt: x.as_array(d, dt),
}


class TestALNCallbacks(unittest.TestCase):
    SIGMA_TEST = 3.2
    MU_TEST = 1.7
    INTERP_EXPECTED = (37, 117, 0.8000000000000185, 0.7875000000002501)
    FIRING_RATE_EXPECTED = 0.09444942503533124
    VOLTAGE_EXPECTED = -56.70455755705249
    TAU_EXPECTED = 0.4487499999999963

    @classmethod
    def setUpClass(cls):
        cls.mass = ExcitatoryALNMass()

    def test_get_interpolation_values(self):
        self.assertTrue(callable(_get_interpolation_values))
        print(type(_get_interpolation_values))
        self.assertTrue(isinstance(_get_interpolation_values, numba.core.registry.CPUDispatcher))
        interp_result = _get_interpolation_values(
            self.SIGMA_TEST,
            self.MU_TEST,
            self.mass.sigma_range,
            self.mass.mu_range,
            self.mass.d_sigma,
            self.mass.d_mu,
        )
        self.assertTupleEqual(interp_result, self.INTERP_EXPECTED)

    def test_table_lookup(self):
        self.assertTrue(callable(_table_lookup))
        self.assertTrue(isinstance(_table_lookup, numba.core.registry.CPUDispatcher))
        firing_rate = _table_lookup(
            self.SIGMA_TEST,
            self.MU_TEST,
            self.mass.sigma_range,
            self.mass.mu_range,
            self.mass.d_sigma,
            self.mass.d_mu,
            self.mass.firing_rate_transfer_function,
        )
        self.assertEqual(firing_rate, self.FIRING_RATE_EXPECTED)

        voltage = _table_lookup(
            self.SIGMA_TEST,
            self.MU_TEST,
            self.mass.sigma_range,
            self.mass.mu_range,
            self.mass.d_sigma,
            self.mass.d_mu,
            self.mass.voltage_transfer_function,
        )
        self.assertEqual(voltage, self.VOLTAGE_EXPECTED)

        tau = _table_lookup(
            self.SIGMA_TEST,
            self.MU_TEST,
            self.mass.sigma_range,
            self.mass.mu_range,
            self.mass.d_sigma,
            self.mass.d_mu,
            self.mass.tau_transfer_function,
        )
        self.assertEqual(tau, self.TAU_EXPECTED)


class ALNMassTestCase(unittest.TestCase):
    def _run_node(self, node, duration, dt):
        coupling_variables = {k: 0.0 for k in node.required_couplings}
        noise = ZeroInput(n=node.num_noise_variables).as_cubic_splines(duration, dt)
        system = jitcdde_input(
            node._derivatives(coupling_variables),
            input=noise,
            callback_functions=node._callbacks(),
        )
        system.constant_past(np.array(node.initial_state))
        system.adjust_diff()
        times = np.arange(dt, duration + dt, dt)
        return np.vstack([system.integrate(time) for time in times])


class TestALNMass(ALNMassTestCase):
    def _create_exc_mass(self):
        exc = ExcitatoryALNMass()
        exc.index = 0
        exc.idx_state_var = 0
        exc.init_mass()
        return exc

    def _create_inh_mass(self):
        inh = InhibitoryALNMass()
        inh.index = 0
        inh.idx_state_var = 0
        inh.init_mass()
        return inh

    def test_init(self):
        aln_exc = self._create_exc_mass()
        aln_inh = self._create_inh_mass()
        self.assertTrue(isinstance(aln_exc, ExcitatoryALNMass))
        self.assertTrue(isinstance(aln_inh, InhibitoryALNMass))
        self.assertDictEqual(_strip_keys(aln_exc.params), _strip_keys(ALN_EXC_DEFAULT_PARAMS))
        self.assertDictEqual(_strip_keys(aln_inh.params), _strip_keys(ALN_INH_DEFAULT_PARAMS))
        # test cascade
        np.testing.assert_equal(aln_exc.mu_range, aln_inh.mu_range)
        np.testing.assert_equal(aln_exc.sigma_range, aln_inh.sigma_range)
        np.testing.assert_equal(aln_exc.firing_rate_transfer_function, aln_inh.firing_rate_transfer_function)
        np.testing.assert_equal(aln_exc.voltage_transfer_function, aln_inh.voltage_transfer_function)
        np.testing.assert_equal(aln_exc.tau_transfer_function, aln_inh.tau_transfer_function)
        for aln in [aln_exc, aln_inh]:
            # test cascade
            self.assertTrue(callable(getattr(aln, "firing_rate_lookup")))
            self.assertTrue(callable(getattr(aln, "voltage_lookup")))
            self.assertTrue(callable(getattr(aln, "tau_lookup")))
            # test callbacks
            self.assertEqual(len(aln._callbacks()), 3)
            self.assertTrue(all(len(callback) == 3 for callback in aln._callbacks()))
            # test numba callbacks
            self.assertEqual(len(aln._numba_callbacks()), 3)
            for numba_callbacks in aln._numba_callbacks():
                self.assertEqual(len(numba_callbacks), 2)
                self.assertTrue(isinstance(numba_callbacks[0], str))
                self.assertTrue(isinstance(numba_callbacks[1], numba.core.registry.CPUDispatcher))
            # test derivatives
            coupling_variables = {k: 0.0 for k in aln.required_couplings}
            self.assertEqual(
                len(aln._derivatives(coupling_variables)),
                aln.num_state_variables,
            )
            self.assertEqual(len(aln.initial_state), aln.num_state_variables)
            self.assertEqual(len(aln.noise_input_idx), aln.num_noise_variables)

    def test_run(self):
        aln_exc = self._create_exc_mass()
        aln_inh = self._create_inh_mass()
        for aln in [aln_exc, aln_inh]:
            result = self._run_node(aln, DURATION, DT)
            self.assertTrue(isinstance(result, np.ndarray))
            self.assertTupleEqual(result.shape, (int(DURATION / DT), aln.num_state_variables))


class TestALNNode(unittest.TestCase):
    def _create_node(self):
        node = ALNNode(exc_seed=SEED, inh_seed=SEED)
        node.index = 0
        node.idx_state_var = 0
        node.init_node()
        return node

    def test_init(self):
        aln = self._create_node()
        self.assertTrue(isinstance(aln, ALNNode))
        self.assertEqual(len(aln), 2)
        self.assertDictEqual(_strip_keys(aln[0].params), _strip_keys(ALN_EXC_DEFAULT_PARAMS))
        self.assertDictEqual(_strip_keys(aln[1].params), _strip_keys(ALN_INH_DEFAULT_PARAMS))
        self.assertTrue(hasattr(aln, "_rescale_connectivity"))
        self.assertEqual(len(aln._sync()), 4 * len(aln))
        self.assertEqual(len(aln.default_network_coupling), 2)
        np.testing.assert_equal(
            np.array(sum([alnm.initial_state for alnm in aln], [])),
            aln.initial_state,
        )

    def test_update_rescale_params(self):
        aln = self._create_node()
        # update connectivity and check rescaling
        old_rescaled = aln.connectivity.copy()
        aln.update_params({"local_connectivity": 2 * ALN_NODE_DEFAULT_CONNECTIVITY})
        np.testing.assert_equal(aln.connectivity, 2 * old_rescaled)

    def test_run(self):
        aln = self._create_node()
        all_results = []
        for backend, noise_func in BACKENDS_TO_TEST.items():
            result = aln.run(
                DURATION, DT, noise_func(ZeroInput(aln.num_noise_variables), DURATION, DT), backend=backend
            )
            self.assertTrue(isinstance(result, xr.Dataset))
            self.assertEqual(len(result), aln.num_state_variables)
            self.assertTrue(all(state_var in result for state_var in aln.state_variable_names[0]))
            self.assertTrue(
                all(result[state_var].shape == (int(DURATION / DT), 1) for state_var in aln.state_variable_names[0])
            )
            all_results.append(result)
        # test results are the same from different backends
        for state_var in all_results[0]:
            corr_mat = np.corrcoef(
                np.vstack([result[state_var].values.flatten().astype(float) for result in all_results])
            )
            self.assertTrue(np.greater(corr_mat, CORR_THRESHOLD).all())

    def test_compare_w_neurolib_native_model(self):
        """
        Compare with neurolib's native ALN model.
        """
        # run this model
        aln_multi = self._create_node()
        multi_result = aln_multi.run(
            DURATION, DT, ZeroInput(aln_multi.num_noise_variables).as_array(DURATION, DT), backend="numba"
        )
        # run neurolib's model
        aln_neurolib = ALNModel(seed=SEED)
        aln_neurolib.params["duration"] = DURATION
        aln_neurolib.params["dt"] = DT
        aln_neurolib.params["mue_ext_mean"] = 0.0
        aln_neurolib.params["mui_ext_mean"] = 0.0
        aln_neurolib.run()
        for (var_multi, var_neurolib) in NEUROLIB_VARIABLES_TO_TEST:
            corr_mat = np.corrcoef(aln_neurolib[var_neurolib], multi_result[var_multi].values.T)
            self.assertTrue(np.greater(corr_mat, CORR_THRESHOLD).all())


class TestALNNetwork(unittest.TestCase):
    SC = np.random.rand(2, 2)
    np.fill_diagonal(SC, 0.0)
    DELAYS = np.array([[0.0, 7.8], [7.8, 0.0]])

    def test_init(self):
        aln = ALNNetwork(self.SC, self.DELAYS)
        self.assertTrue(isinstance(aln, ALNNetwork))
        self.assertEqual(len(aln), self.SC.shape[0])
        self.assertEqual(aln.initial_state.shape[0], aln.num_state_variables)
        self.assertEqual(aln.default_output, f"r_mean_{EXC}")

    def test_run(self):
        aln = ALNNetwork(self.SC, self.DELAYS, exc_seed=SEED, inh_seed=SEED)
        all_results = []
        for backend, noise_func in BACKENDS_TO_TEST.items():
            result = aln.run(
                DURATION,
                DT,
                noise_func(ZeroInput(aln.num_noise_variables), DURATION, DT),
                backend=backend,
            )
            self.assertTrue(isinstance(result, xr.Dataset))
            self.assertEqual(len(result), aln.num_state_variables / aln.num_nodes)
            self.assertTrue(all(result[result_].shape == (int(DURATION / DT), aln.num_nodes) for result_ in result))
            all_results.append(result)
        # test results are the same from different backends
        for state_var in all_results[0]:
            all_ts = np.vstack([result[state_var].values.flatten().astype(float) for result in all_results])
            if np.isnan(all_ts).any():
                continue
            corr_mat = np.corrcoef(all_ts)
            print(state_var, corr_mat, np.var(all_ts))
            self.assertTrue(np.greater(corr_mat, CORR_THRESHOLD).all())

    @pytest.mark.xfail
    def test_compare_w_neurolib_native_model(self):
        """
        Compare with neurolib's native ALN model.
        Marked with xfail, since sometimes fail on specific python version on
        Linux, no idea why, but the model works...
        """
        aln_multi = ALNNetwork(self.SC, self.DELAYS, exc_seed=SEED, inh_seed=SEED)
        multi_result = aln_multi.run(
            DURATION, DT, ZeroInput(aln_multi.num_noise_variables).as_array(DURATION, DT), backend="numba"
        )
        # run neurolib's model
        aln_neurolib = ALNModel(Cmat=self.SC, Dmat=self.DELAYS, seed=SEED)
        aln_neurolib.params["duration"] = DURATION
        aln_neurolib.params["dt"] = DT
        # there is no "global coupling" parameter in MultiModel
        aln_neurolib.params["K_gl"] = 1.0
        # delays <-> length matrix
        aln_neurolib.params["signalV"] = 1.0
        aln_neurolib.params["sigma_ou"] = 0.0
        aln_neurolib.params["mue_ext_mean"] = 0.0
        aln_neurolib.params["mui_ext_mean"] = 0.0
        # match initial state at least for current - this seems to be enough
        aln_neurolib.params["mufe_init"] = np.array(
            [aln_multi[0][0].initial_state[0], aln_multi[1][0].initial_state[0]]
        )
        aln_neurolib.params["mufi_init"] = np.array(
            [aln_multi[0][1].initial_state[0], aln_multi[1][1].initial_state[0]]
        )
        aln_neurolib.run()
        for (var_multi, var_neurolib) in NEUROLIB_VARIABLES_TO_TEST:
            for node_idx in range(len(aln_multi)):
                corr_mat = np.corrcoef(
                    aln_neurolib[var_neurolib][node_idx, :], multi_result[var_multi].values.T[node_idx, :]
                )
                print(corr_mat)
                self.assertTrue(np.greater(corr_mat, CORR_THRESHOLD).all())


if __name__ == "__main__":
    unittest.main()
